# -*- coding: utf-8 -*-
#
# Copyright (c) 2018 hankei6km
# Licensed under the MIT License. See LICENSE.txt in the project root.

from xrosfs import main

if __name__ == '__main__':
    import sys
    main(argv=sys.argv[1:])
